//#include <stdio.h>
//#include <stdlib.h>

#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/cmn_err.h>
#include <sys/modctl.h>
#include <sys/conf.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>
#include <sys/sunldi.h>

#include "myzil.h"

typedef struct blk_state {
	ldi_handle_t lh;
	ldi_ident_t li;
	dev_info_t *dip;
	minor_t minor;
	int flags;
    char *data;
	kmutex_t lock;
	uint64_t Nblocks;
	struct dev_ops *ldev;
} blk_state_t;

#define BLK_OPENED 0x1 /* lh is valid */
#define BLK_IDENTED 0x2 /* li is valid */

static int blk_info(dev_info_t *, ddi_info_cmd_t, void *, void **);
static int blk_attach(dev_info_t *, ddi_attach_cmd_t);
static int blk_detach(dev_info_t *, ddi_detach_cmd_t);
static int blk_open(dev_t *, int, int, cred_t *);
static int blk_close(dev_t, int, int, cred_t *);
static int blk_write(dev_t, struct uio *, cred_t *);
static int blk_read(dev_t, struct uio *, cred_t *);
static int blk_strategy(struct buf *);
static int blk_prop_op(dev_t, dev_info_t *, ddi_prop_op_t,
					   int, char *, caddr_t, int *);
static int blk_ioctl(dev_t, int, intptr_t, int, cred_t *, int *);

static void *blk_statep;

static struct cb_ops blk_cb_ops = {
	blk_open, /* open */
	blk_close, /* close */
	nodev, /* strategy */
	nodev, /* print */
	nodev, /* dump */
	nodev, /* read */
	nodev, /* write */
	blk_ioctl, /* ioctl */
	nodev, /* devmap */
	nodev, /* mmap */
	nodev, /* segmap */
	nochpoll, /* poll */
	blk_prop_op, /* prop_op */
	NULL, /* streamtab */
	D_NEW | D_MP, /* cb_flag */
	CB_REV, /* cb_rev */
	nodev, /* aread */
	nodev /* awrite */
};

static struct dev_ops blk_dev_ops = {
	DEVO_REV, /* devo_rev, */
	0, /* refcnt */
	blk_info, /* getinfo */
	nulldev, /* identify */
	nulldev, /* probe */
	blk_attach, /* attach */
	blk_detach, /* detach */
	nodev, /* reset */
	&blk_cb_ops, /* cb_ops */
	NULL, /* bus_ops */
	NULL /* power */
};

static struct modldrv modldrv = {
	&mod_driverops,
	"LDI character driver",
	&blk_dev_ops
};

static struct modlinkage modlinkage = {
	MODREV_1,
	&modldrv,
	NULL
};

int
_init(void)
{
	int rv;
    ldi_handle_t  lhandle;
    ldi_ident_t ident;

	if ((rv = ddi_soft_state_init(&blk_statep, sizeof (blk_state_t),
					0)) != 0) {
		cmn_err(CE_WARN, "blk _init: soft state init failed\n");
		return (rv);
	}
	if ((rv = mod_install(&modlinkage)) != 0) {
		cmn_err(CE_WARN, "blk _init: mod_install failed\n");
		goto FAIL;
	}
	return (rv);
	/*NOTEREACHED*/

FAIL:
	ddi_soft_state_fini(&blk_statep);
	return (rv);
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}

int
_fini(void)
{
	int rv;
	if ((rv = mod_remove(&modlinkage)) != 0) {
		return(rv);
	}
	ddi_soft_state_fini(&blk_statep);
	return (rv);
}

/*
* 1:1 mapping between minor number and instance
*/
static int
blk_info(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg, void **result)
{
	int inst;
	minor_t minor;
	blk_state_t *statep;
	char *myname = "blk_info";
	minor = getminor((dev_t)arg);
	inst = minor;
	switch (infocmd) {
		case DDI_INFO_DEVT2DEVINFO:
			statep = ddi_get_soft_state(blk_statep, inst);
			if (statep == NULL) {
				cmn_err(CE_WARN, "%s: get soft state "
						"failed on inst %d\n", myname, inst);
				return (DDI_FAILURE);
			}
			*result = (void *)statep->dip;
			break;
		case DDI_INFO_DEVT2INSTANCE:
			*result = (void *)inst;
			break;
		default:
			break;
	}
	return (DDI_SUCCESS);
}

static int
blk_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int inst,fd, maj_number;
	blk_state_t *statep;
	char *myname = "blk_attach";
	struct vnode *vp;
	switch (cmd) {
		case DDI_ATTACH:
			inst = ddi_get_instance(dip);
			if (ddi_soft_state_zalloc(blk_statep, inst) != DDI_SUCCESS) {
				cmn_err(CE_WARN, "%s: ddi_soft_state_zallac failed "
						"on inst %d\n", myname, inst);
				goto FAIL;
			}
			statep = (blk_state_t *)ddi_get_soft_state(blk_statep, inst);
			if (statep == NULL) {
				cmn_err(CE_WARN, "%s: ddi_get_soft_state failed on "
						"inst %d\n", myname, inst);
				goto FAIL;
			}
			statep->dip = dip;
			statep->minor = inst;
			if (ddi_create_minor_node(dip, "node", S_IFCHR, 0,
						NULL, NULL) != DDI_SUCCESS) {
						#if 0
			if (ddi_create_minor_node(dip, "node", S_IFCHR, statep->minor,
						DDI_NT_BLOCK, 0) != DDI_SUCCESS) {
						#endif
				cmn_err(CE_WARN, "%s: ddi_create_minor_node failed on "
						"inst %d\n", myname, inst);
				goto FAIL;
			}
            
            //statep->data = vmalloc();

			mutex_init(&statep->lock, NULL, MUTEX_DRIVER, NULL);
			
			return (DDI_SUCCESS);
		case DDI_RESUME:
		case DDI_PM_RESUME:
		default:
			break;
	}
	return (DDI_FAILURE);
	/*NOTREACHED*/
FAIL:
	ddi_soft_state_free(blk_statep, inst);
	ddi_remove_minor_node(dip, NULL);
	return (DDI_FAILURE);
}

static int
blk_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	int inst;
	blk_state_t *statep;
	char *myname = "blk_detach";
	inst = ddi_get_instance(dip);
	statep = ddi_get_soft_state(blk_statep, inst);
	if (statep == NULL) {
		cmn_err(CE_WARN, "%s: get soft state failed on "
				"inst %d\n", myname, inst);
		return (DDI_FAILURE);
	}
	if (statep->dip != dip) {
		cmn_err(CE_WARN, "%s: soft state does not match devinfo "
				"on inst %d\n", myname, inst);
		return (DDI_FAILURE);
	}
	switch (cmd) {
		case DDI_DETACH:
			mutex_destroy(&statep->lock);
			ddi_soft_state_free(blk_statep, inst);
			ddi_remove_minor_node(dip, NULL);
			return (DDI_SUCCESS);
		case DDI_SUSPEND:
		case DDI_PM_SUSPEND:
		default:
			break;
	}
	return (DDI_FAILURE);
}

/*
* on this driver's open, we open the target specified by a 
* property and store
* the layered handle and ident in our soft state. a good target would be
* "/dev/console" or more interestingly, a pseudo terminal as specified by the
* tty command
*/
/*ARGSUSED*/
static int
blk_open(dev_t *devtp, int oflag, int otyp, cred_t *credp)
{
	int rv, inst = getminor(*devtp);
	blk_state_t *statep;
	char *myname = "blk_open";
	dev_info_t *dip;
    char *blk_targ = "/dev/dsk/c3d1s13";
	
    statep = (blk_state_t *)ddi_get_soft_state(blk_statep, inst);
	if (statep == NULL) {
		cmn_err(CE_WARN, "%s: ddi_get_soft_state failed on inst %d\n",
				myname, inst);
		return (EIO);
	}
	dip = statep->dip;
	
	//cmn_err(CE_WARN, "*** blk_target: %s ***\n", blk_targ);
    
	mutex_enter(&statep->lock);

	rv = ldi_ident_from_dev(*devtp, &statep->li);
	if (rv != 0) {
		cmn_err(CE_WARN, "%s: ldi_ident_from_dev failed on inst %d\n",
				myname, inst);
		goto FAIL;
	}

	statep->flags |= BLK_IDENTED;
	rv = ldi_open_by_name(blk_targ, FREAD, 
						  credp, &statep->lh, statep->li);
	if (rv != 0) {
		cmn_err(CE_WARN, "%s: ldi_open_by_name failed on inst %d\n",
				myname, inst);
		goto FAIL;
	}

	statep->flags |= BLK_OPENED;
	cmn_err(CE_CONT, "\n%s: opened target %s successfully on inst %d\n",
			myname, blk_targ, inst);

	rv = 0;

	//cmn_err(CE_WARN, "%s: Success ! \n", myname);
			
FAIL:
	/* cleanup on error */
	if (rv != 0) {
		if (statep->flags & BLK_OPENED)
			(void)ldi_close(statep->lh, FREAD, credp);
		if (statep->flags & BLK_IDENTED)
			ldi_ident_release(statep->li);
		statep->flags &= ~(BLK_OPENED | BLK_IDENTED);
	}

	mutex_exit(&statep->lock);
	return (rv);
}

/*
* on this driver's close, we close the target indicated by the lh member
* in our soft state and release the ident, li as well. in fact, we MUST do
* both of these at all times even if close yields an error because the
* device framework effectively closes the device, releasing all data
* associated with it and simply returning whatever value the target’s
* close(9E) returned. therefore, we must as well.
*/
/*ARGSUSED*/
static int
blk_close(dev_t devt, int oflag, int otyp, cred_t *credp)
{
	int rv, inst = getminor(devt);
	blk_state_t *statep;
	char *myname = "blk_close";
	statep = (blk_state_t *)ddi_get_soft_state(blk_statep, inst);
	if (statep == NULL) {
		cmn_err(CE_WARN, "%s: ddi_get_soft_state failed on "
				"inst %d\n", myname, inst);
		return (EIO);
	}
	mutex_enter(&statep->lock);
	rv = ldi_close(statep->lh, FREAD, credp);
	if (rv != 0) {
		cmn_err(CE_WARN, "%s: ldi_close failed on inst %d, but will ",
				"continue to release ident\n", myname, inst);
	}
	ldi_ident_release(statep->li);
	if (rv == 0) {
		cmn_err(CE_CONT, "\n%s: closed target successfully on "
				"inst %d\n", myname, inst);
	}
	statep->flags &= ~(BLK_OPENED | BLK_IDENTED);
	mutex_exit(&statep->lock);
	return (rv);
}

static int
blk_prop_op(dev_t devt, dev_info_t *dip, ddi_prop_op_t prop_op,
			int flags, char *name, caddr_t valuep, int *lengthp)
{
	int rv, inst = getminor(devt);
	blk_state_t *statep;
	char *myname = "blk_prop_op";
	
	statep = (blk_state_t *)ddi_get_soft_state(blk_statep, inst);
	if (statep == NULL) {
		cmn_err(CE_WARN, "%s: ddi_get_soft_state failed on inst %d\n",
				myname, inst);
		return (EIO);
	}
	return(ldi_prop_op(statep->lh, prop_op, flags, name, valuep, lengthp)); 
}

static int 
blk_ioctl(dev_t devt, int cmd, intptr_t arg, int mode,
		  cred_t *credp, int *rvalp)
{
	int rv, inst = getminor(devt);
	blk_state_t *statep;
	char *myname = "blk_ioctl";
	
	statep = (blk_state_t *)ddi_get_soft_state(blk_statep, inst);
	if (statep == NULL) {
		cmn_err(CE_WARN, "%s: ddi_get_soft_state failed on inst %d\n",
				myname, inst);
		return (EIO);
	}
	return(ldi_ioctl(statep->lh, cmd, arg, mode, credp, rvalp)); 
}
