#include <sys/types.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/open.h>
#include <sys/cred.h>
#include <sys/cmn_err.h>
#include <sys/modctl.h>
#include <sys/conf.h>
#include <sys/stat.h>
#include <sys/ddi.h>
#include <sys/vnode.h>
#include <sys/sunddi.h>
#include <sys/sunldi.h>

#include "myzil.h"
#include "stats.h"

typedef struct lyr_state {
	ldi_handle_t lh;
	ldi_ident_t li;
	dev_info_t *dip;
	minor_t minor;
	int flags;
	int nopened;
	kmutex_t lock;
	uint64_t Nblocks;

	/* This is used for ioctl counters */
    int start;
    uint64_t dblocks;	/* data blocks */
    uint64_t jblocks;	/* journal blocks */
    uint64_t oblocks;	/* other blocks */
    uint64_t ublocks;	/* uberblocks */ 
	
	/* This is to see when the versions are created */
	timestruc_t curr_time;	/* start a new version */
	timestruc_t end_time;	/* end time of a new  version */
	int uber_blk_cnt;		/* how many uberblocks were written */
	int is_start;			/* is it start or end time ? */

	struct vnode *log_vp;    /* File handle to our log */
	offset_t log_offset;     /* offset to write for our log */
#ifdef DEBUG_STATS
    struct vnode *debug_vp;    /* File handle to our log */
	offset_t debug_offset;     /* offset to write for our log */
#endif
    char log_buf[4096];           /* buffer to write the log records */

    struct dev_ops *ldev;

} lyr_state_t;

#define LYR_OPENED 0x1 /* lh is valid */
#define LYR_IDENTED 0x2 /* li is valid */

static int lyr_info(dev_info_t *, ddi_info_cmd_t, void *, void **);
static int lyr_attach(dev_info_t *, ddi_attach_cmd_t);
static int lyr_detach(dev_info_t *, ddi_detach_cmd_t);
static int lyr_open(dev_t *, int, int, cred_t *);
static int lyr_close(dev_t, int, int, cred_t *);
static int lyr_write(dev_t, struct uio *, cred_t *);
static int lyr_read(dev_t, struct uio *, cred_t *);
static int lyr_strategy(struct buf *);
static int lyr_prop_op(dev_t, dev_info_t *, ddi_prop_op_t,
					   int, char *, caddr_t, int *);
static int lyr_ioctl(dev_t, int, intptr_t, int, cred_t *, int *);
void lyr_reset_counters(lyr_state_t *statep);

static void *lyr_statep;

static struct cb_ops lyr_cb_ops = {
	lyr_open, /* open */
	lyr_close, /* close */
	lyr_strategy, /* strategy */
	nodev, /* print */
	nodev, /* dump */
	nodev, /* read */
	nodev, /* write */
	lyr_ioctl, /* ioctl */
	nodev, /* devmap */
	nodev, /* mmap */
	nodev, /* segmap */
	nochpoll, /* poll */
	lyr_prop_op, /* prop_op */
	NULL, /* streamtab */
	D_NEW | D_MP, /* cb_flag */
	CB_REV, /* cb_rev */
	nodev, /* aread */
	nodev /* awrite */
};

static struct dev_ops lyr_dev_ops = {
	DEVO_REV, /* devo_rev, */
	0, /* refcnt */
	lyr_info, /* getinfo */
	nulldev, /* identify */
	nulldev, /* probe */
	lyr_attach, /* attach */
	lyr_detach, /* detach */
	nodev, /* reset */
	&lyr_cb_ops, /* cb_ops */
	NULL, /* bus_ops */
	NULL /* power */
};

static struct modldrv modldrv = {
	&mod_driverops,
	"LDI block driver",
	&lyr_dev_ops
};

static struct modlinkage modlinkage = {
	MODREV_1,
	&modldrv,
	NULL
};

int
_init(void)
{
	int rv, Errno;
	lyr_state_t *ls;


	if ((rv = ddi_soft_state_init(&lyr_statep, sizeof (lyr_state_t),
					0)) != 0) {
		cmn_err(CE_WARN, "lyr _init: soft state init failed\n");
		return (rv);
	}
	if ((rv = mod_install(&modlinkage)) != 0) {
		cmn_err(CE_WARN, "lyr _init: mod_install failed\n");
		goto FAIL;
	}
	ls = (lyr_state_t *)lyr_statep;
	//lyr_statep.nopened = 0;
	ls->start = 0;
	lyr_reset_counters(ls);
#if 0
    ls->log_buf = kmem_alloc(LYR_FS_BLK_SIZE, KM_SLEEP);
    
    if (!ls->log_buf) {
	    mod_remove(&modlinkage);
		cmn_err(CE_WARN, "lyr kmem_alloc failed \n");
        rv = -1;
        goto FAIL;
    } 
#endif
	return (rv);
	/*NOTEREACHED*/

FAIL:
	ddi_soft_state_fini(&lyr_statep);
	return (rv);
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}

int
_fini(void)
{
	int rv;
    lyr_state_t *statep = (lyr_state_t *) lyr_statep;

	if ((rv = mod_remove(&modlinkage)) != 0) {
		return(rv);
	}

   // kmem_free(statep->log_buf, LYR_FS_BLK_SIZE);

	ddi_soft_state_fini(&lyr_statep);
	return (rv);
}

/*
* 1:1 mapping between minor number and instance
*/
static int
lyr_info(dev_info_t *dip, ddi_info_cmd_t infocmd, void *arg, void **result)
{
	int inst;
	minor_t minor;
	lyr_state_t *statep;
	char *myname = "lyr_info";
	minor = getminor((dev_t)arg);
	inst = minor;
	switch (infocmd) {
		case DDI_INFO_DEVT2DEVINFO:
			statep = ddi_get_soft_state(lyr_statep, inst);
			if (statep == NULL) {
				cmn_err(CE_WARN, "%s: get soft state "
						"failed on inst %d\n", myname, inst);
				return (DDI_FAILURE);
			}
			*result = (void *)statep->dip;
			break;
		case DDI_INFO_DEVT2INSTANCE:
			*result = (void *)inst;
			break;
		default:
			break;
	}
	return (DDI_SUCCESS);
}

static int
lyr_attach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int inst,fd;
	lyr_state_t *statep;
	char *myname = "lyr_attach";
	struct vnode *vp;
	switch (cmd) {
		case DDI_ATTACH:
			inst = ddi_get_instance(dip);
			if (ddi_soft_state_zalloc(lyr_statep, inst) != DDI_SUCCESS) {
				cmn_err(CE_WARN, "%s: ddi_soft_state_zallac failed "
						"on inst %d\n", myname, inst);
				goto FAIL;
			}
			statep = (lyr_state_t *)ddi_get_soft_state(lyr_statep, inst);
			if (statep == NULL) {
				cmn_err(CE_WARN, "%s: ddi_get_soft_state failed on "
						"inst %d\n", myname, inst);
				goto FAIL;
			}
			statep->dip = dip;
			statep->minor = inst;
			statep->nopened = 0;
			if (ddi_create_minor_node(dip, "node", S_IFBLK, statep->minor,
						DDI_NT_BLOCK, 0) != DDI_SUCCESS) {
				cmn_err(CE_WARN, "%s: ddi_create_minor_node failed on "
						"inst %d\n", myname, inst);
				goto FAIL;
			}
			mutex_init(&statep->lock, NULL, MUTEX_DRIVER, NULL);
			
			return (DDI_SUCCESS);
		case DDI_RESUME:
		case DDI_PM_RESUME:
		default:
			break;
	}
	return (DDI_FAILURE);
	/*NOTREACHED*/
FAIL:
	ddi_soft_state_free(lyr_statep, inst);
	ddi_remove_minor_node(dip, NULL);
	return (DDI_FAILURE);
}

static int
lyr_detach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	int inst;
	lyr_state_t *statep;
	char *myname = "lyr_detach";
	inst = ddi_get_instance(dip);
	statep = ddi_get_soft_state(lyr_statep, inst);
	if (statep == NULL) {
		cmn_err(CE_WARN, "%s: get soft state failed on "
				"inst %d\n", myname, inst);
		return (DDI_FAILURE);
	}
	if (statep->dip != dip) {
		cmn_err(CE_WARN, "%s: soft state does not match devinfo "
				"on inst %d\n", myname, inst);
		return (DDI_FAILURE);
	}
	switch (cmd) {
		case DDI_DETACH:
			mutex_destroy(&statep->lock);
			ddi_soft_state_free(lyr_statep, inst);
			ddi_remove_minor_node(dip, NULL);
			return (DDI_SUCCESS);
		case DDI_SUSPEND:
		case DDI_PM_SUSPEND:
		default:
			break;
	}
	return (DDI_FAILURE);
}

/*
* on this driver's open, we open the target specified by a 
* property and store
* the layered handle and ident in our soft state. a good target would be
* "/dev/console" or more interestingly, a pseudo terminal as specified by the
* tty command
*/
/*ARGSUSED*/
static int
lyr_open(dev_t *devtp, int oflag, int otyp, cred_t *credp)
{
	int rv, inst = getminor(*devtp);
	lyr_state_t *statep;
	char *myname = "lyr_open";
	dev_info_t *dip;
	char *lyr_targ = NULL;
	
	//cmn_err(CE_WARN,"oflag: %d otype:%d\n",oflag, otyp);
	if (oflag == 1073741825) {
		//cmn_err(CE_WARN, "OPENED EARLY\n");
		return 0;
	}

	statep = (lyr_state_t *)ddi_get_soft_state(lyr_statep, inst);
	if (statep == NULL) {
		cmn_err(CE_WARN, "%s: ddi_get_soft_state failed on inst %d\n",
				myname, inst);
		return (EIO);
	}
	dip = statep->dip;
	/*
	 * our target device to open should be specified by the "lyr_targ"
	 * string property, which should be set in this driver's .conf file
	 */
	if (ddi_prop_lookup_string(DDI_DEV_T_ANY, dip, DDI_PROP_NOTPROM,
				"lyr_targ", &lyr_targ) != DDI_PROP_SUCCESS) {
		cmn_err(CE_WARN, "%s: ddi_prop_lookup_string failed on "
				"inst %d %s\n", myname, inst, lyr_targ);
		return (EIO);
	}
	
	//cmn_err(CE_WARN, "*** lyr_target: %s %lu ***\n", lyr_targ, *devtp);

	/*
	 * since we only have one pair of lhs and lis available, we dont
	 * allow multiple on the same instance
	 */
	mutex_enter(&statep->lock);
	if (statep->flags & (LYR_OPENED | LYR_IDENTED)) {
		cmn_err(CE_WARN, "%s: multiple layered opens or idents "
				"from inst %d not allowed\n", myname, inst);
			mutex_exit(&statep->lock);
			ddi_prop_free(lyr_targ);
			return (EIO);
	}

	rv = ldi_ident_from_dev(*devtp, &statep->li);
	if (rv != 0) {
		cmn_err(CE_WARN, "%s: ldi_ident_from_dev failed on inst %d\n",
				myname, inst);
		goto FAIL;
	}
	statep->flags |= LYR_IDENTED;
	rv = ldi_open_by_name(lyr_targ, FREAD | FWRITE, credp, &statep->lh,
			statep->li);
	if (rv != 0) {
		cmn_err(CE_WARN, "%s: ldi_open_by_name failed on inst %d\n",
				myname, inst);
		goto FAIL;
	}
	statep->flags |= LYR_OPENED;
	cmn_err(CE_CONT, "\n%s: opened target %s successfully on inst %d\n",
			myname, lyr_targ, inst);

	rv = 0;
	if(ldi_get_size(statep->lh, &statep->Nblocks) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "%s: ldi_get_size failed on "
				"inst %d\n", myname, inst);
		goto FAIL;
	}
FAIL:
	/* cleanup on error */
	if (rv != 0) {
		if (statep->flags & LYR_OPENED)
			(void)ldi_close(statep->lh, FREAD | FWRITE, credp);
		if (statep->flags & LYR_IDENTED)
			ldi_ident_release(statep->li);
		statep->flags &= ~(LYR_OPENED | LYR_IDENTED);
	}
	mutex_exit(&statep->lock);
	if (lyr_targ != NULL)
		ddi_prop_free(lyr_targ);
	return (rv);
}

/*
* on this driver's close, we close the target indicated by the lh member
* in our soft state and release the ident, li as well. in fact, we MUST do
* both of these at all times even if close yields an error because the
* device framework effectively closes the device, releasing all data
* associated with it and simply returning whatever value the target’s
* close(9E) returned. therefore, we must as well.
*/
/*ARGSUSED*/
static int
lyr_close(dev_t devt, int oflag, int otyp, cred_t *credp)
{
	int rv, inst = getminor(devt);
	lyr_state_t *statep;
	char *myname = "lyr_close";
	
	//cmn_err(CE_WARN,"close: oflag: %d otype:%d\n",oflag, otyp);
	if (oflag == 1073741825) {
		//cmn_err(CE_WARN, "CLOSED EARLY\n");
		return 0;
	}
	statep = (lyr_state_t *)ddi_get_soft_state(lyr_statep, inst);
	if (statep == NULL) {
		cmn_err(CE_WARN, "%s: ddi_get_soft_state failed on "
				"inst %d\n", myname, inst);
		return (EIO);
	}

	mutex_enter(&statep->lock);
	rv = ldi_close(statep->lh, FREAD | FWRITE, credp);
	if (rv != 0) {
		cmn_err(CE_WARN, "%s: ldi_close failed on inst %d, but will ",
				"continue to release ident\n", myname, inst);
	}
	ldi_ident_release(statep->li);
	if (rv == 0) {
		cmn_err(CE_CONT, "\n%s: closed target successfully on "
				"inst %d\n", myname, inst);
	}
	statep->flags &= ~(LYR_OPENED | LYR_IDENTED);
	mutex_exit(&statep->lock);
	return (rv);
}

int check_ubber_block(struct buf *bp)
{
	uberblock_t *ub = (uberblock_t *) bp->b_un.b_addr;
	if (ub->ub_magic == UBERBLOCK_MAGIC)
		return 0;

	return -1;
}

int check_for_objset(struct buf *bp)
{
	objset_phys_t *obset = (objset_phys_t *) bp->b_un.b_addr;
	dnode_phys_t *dn = (dnode_phys_t *) bp->b_un.b_addr;
	
	if (obset->os_type > 3)
		return -1;
	if (dn->dn_type > 24)
		return -1;
	if (dn->dn_compress > 3)
		return -1;
	if (dn->dn_checksum > 8)
		return -1;
	if (dn->dn_nblkptr > 3)
		return -1;
	if (dn->dn_bonustype == 4  || dn->dn_bonustype == 7  || 
		dn->dn_bonustype == 12 || dn->dn_bonustype == 16 || 
		dn->dn_bonustype == 17)
		if (dn->dn_bonuslen < 64 || dn->dn_bonuslen > 320)
			return -1;
	return 0;
}

int lyr_zil_process_create(lr_t *lrt) 
{
	lr_create_t *lrc = (lr_create_t *) lrt;
	char *name = (char*)lrc + 
					(lrc->lr_common.lrc_reclen - sizeof(lr_create_t));
	cmn_err(CE_WARN, "name: %c \n", *name);
	return 0;
}

int check_zil_block(struct buf *bp)
{	
	uint64_t max_block_size; 
	char *start;
    
    if (!bp || !bp->b_un.b_addr) {
        cmn_err(CE_WARN, "** ZIL: not a valid bp pointer ***\n"); 
        return -1;
    }

    max_block_size = bp->b_bcount;
	start =	((char *) bp->b_un.b_addr) + max_block_size - 32 - 8;
	
    if (*(uint64_t*)start == ZBT_MAGIC) {
		
        #ifdef DEBUG_ZIL_NO

        lr_t *lrt = (lr_t *) bp->b_un.b_addr;
        uint64_t count = 0, cnt =0;
        uint64_t max_block_size = bp->b_bcount;
        zil_trailer_t *zit;
loop:
        if (max_block_size >= 4096 && 
            lrt->lrc_txtype >= TX_CREATE  &&
            lrt->lrc_txtype < TX_MAX_TYPE &&
            lrt->lrc_reclen) {
            cmn_err(CE_WARN,"b:%llu\tt:%lu g:%lu q:%lu c:%lu cn:%lu rl:%lu bs:%lu \n",
                    bp->b_blkno, lrt->lrc_txtype, lrt->lrc_txg, lrt->lrc_seq, 
                    count, cnt++, lrt->lrc_reclen, max_block_size);
            if (lrt->lrc_txtype  == TX_WRITE) {
                lr_write_t *lrw = (lr_write_t*) lrt;    
                cmn_err(CE_WARN, "write: of:%llu l:%llu \n",
                        lrw->lr_offset, lrw->lr_length);
            }
            if ((lrt->lrc_reclen + count)  < max_block_size) {
                count += lrt->lrc_reclen;
                lrt = (lr_t *) (((char *)bp->b_un.b_addr) + count);
                cmn_err(CE_WARN, "c: %lu s: %lu ty: :%lu rl: %lu\n",
                        count, bp->b_bcount, 
                        lrt->lrc_txtype < TX_MAX_TYPE ? lrt->lrc_txtype : 0,
                        lrt->lrc_txtype < TX_MAX_TYPE &&  
                        lrt->lrc_txtype >= TX_CREATE ? lrt->lrc_reclen : 0);
                goto loop;
            }
        }

        #endif


        #if 0 //DEBUG_ZIL	
		zil_trailer_t *zit;
    	dva_t *dva;
		zit = (zil_trailer_t *) (((char *) bp->b_un.b_addr) + 
							(max_block_size-sizeof(zil_trailer_t)));
		
		dva = &zit->zit_next_blk.blk_dva[0];
        
		cmn_err(CE_WARN, "ZIL b: %lu s:%lu nu: %lu %d \n \
				dev:%lld bno:%llu as:%lld %-5s \n", 
				bp->b_blkno, bp->b_bcount, zit->zit_nused, 
                sizeof(zil_trailer_t), DVA_GET_VDEV(dva), 
				(DVA_GET_OFFSET(dva) >> 9) + 8192, DVA_GET_ASIZE(dva),
				DVA_GET_GANG(dva) ? "TRUE" : "FALSE");
		#endif
        
		return 0;
	}
	else
		return -1;
}

void lyr_print_diff_time(lyr_state_t *statep)
{
	unsigned long int sec, nsec;
	sec = statep->end_time.tv_sec - statep->curr_time.tv_sec;
	nsec = statep->end_time.tv_nsec - statep->curr_time.tv_nsec;
	cmn_err(CE_WARN, "New Version interval : %lu %lu \n", sec, nsec); 
}

void lyr_process_block_data(struct buf *bp, lyr_state_t *statep)
{

    int total_blocks = bp->b_bcount >> 9;
	int count = 0, num_blocks, noinc = 0;
    
	if (bp->b_bcount == 1024) {
		/* check for ubber block */
		if (!check_ubber_block(bp)) {
#ifdef DEBUG_UBER
			if (statep->ublocks % 4 == 0)  {
                size_t bufsize;
				gethrestime(&statep->curr_time);
                bufsize = snprintf(statep->log_buf, LYR_FS_BLK_SIZE,
                    "UBER:\t%llu\t%llu\t%llu\t%llu\t%lu\t%lu\t%llu\t%llu\n", 
                                    statep->oblocks, statep->dblocks, 
                                    statep->jblocks, statep->ublocks/4,
                                    statep->curr_time.tv_sec, 
                                    statep->curr_time.tv_nsec, 
                                    bp->b_blkno, bp->b_bcount);
                lyr_write_log_record(statep, statep->log_buf, bufsize);
            }
#endif
			statep->ublocks++;
			return;
		}
	}

    if (!check_zil_block(bp)) {
#ifdef DEBUG_ZIL
	    gethrestime(&statep->curr_time);
        ssize_t bufsize = snprintf(statep->log_buf, LYR_FS_BLK_SIZE,
                "ZIL:\t%llu\t%llu\t%lu\t%lu\n",
                bp->b_blkno, bp->b_bcount,
                statep->curr_time.tv_sec,
                statep->curr_time.tv_nsec);
                lyr_write_log_record(statep, statep->log_buf, bufsize);    
#endif
        statep->jblocks+= total_blocks;
    }    
	else {
        int data = 0;
		num_blocks = total_blocks;

		while(count < num_blocks) {
   			if(strncmp(((char *)bp->b_un.b_addr)+(count * 512), "$%^&9#@",7) == 0) {
#ifdef DEBUG_DATA_STATS
                int *num;
				gethrestime(&statep->curr_time);
                num = (int *) (((char *)bp->b_un.b_addr)+((count * 512)+8)) ;
                ssize_t bufsize = snprintf(statep->log_buf, LYR_FS_BLK_SIZE,
                                            "%d\t%lu\t%lu\n", *num,
                                            statep->curr_time.tv_sec,
                                            statep->curr_time.tv_nsec);
                lyr_write_debug_record(statep, statep->log_buf, bufsize);    
#endif
#ifdef DEBUG_DATA
                int *num;
				gethrestime(&statep->curr_time);
                num = (int *) (((char *)bp->b_un.b_addr)+((count * 512)+8)) ;
                ssize_t bufsize = snprintf(statep->log_buf, LYR_FS_BLK_SIZE,
                                            "DATA:\t%lu\t%lu\t%d\t%lu\t%lu\n", 
                                            bp->b_blkno, bp->b_bcount, *num,
                                            statep->curr_time.tv_sec,
                                            statep->curr_time.tv_nsec);
                lyr_write_log_record(statep, statep->log_buf, bufsize);    
#endif  
                data = 1;
                statep->dblocks++;
			} else {
                /* Eventhough it doesnt have a data block marker
                 * it part of the wasted datablock space.
                 */
                if (data == 1)
                    statep->dblocks++;
                else
                    statep->oblocks++;
			}
			count++;
		}
	}
    
	return;

#if 0
	else if (!check_for_objset(bp)) {
		/* check for bjectset */
		cmn_err(CE_WARN, "OS b: %lu \n",bp->b_blkno);
		return;
	}
	
	else 
	#endif 
	#if 0
	else
		cmn_err(CE_WARN, "UC b: %lu s: %lu\n", 
				bp->b_blkno, bp->b_bcount);
	#endif
}

void lyr_process_block_data_onread(struct buf *bp, lyr_state_t *statep)
{

    int total_blocks = bp->b_bcount >> 9;
	int count = 0, num_blocks, noinc = 0;
    
    int data = 0;
	num_blocks = total_blocks;

    while(count < num_blocks) {
        if(strncmp(((char *)bp->b_un.b_addr)+(count * 512), "$%^&9#@",7) == 0) {
            #ifdef DEBUG_DATA
            int *num;
            gethrestime(&statep->curr_time);
            num = (int *) (((char *)bp->b_un.b_addr)+((count * 512)+8)) ;
            ssize_t bufsize = snprintf(statep->log_buf, LYR_FS_BLK_SIZE,
                    "RDATA:\t%lu\t%lu\t%d\t%lu\t%lu\n", 
                    bp->b_blkno, bp->b_bcount, *num,
                    statep->curr_time.tv_sec,
                    statep->curr_time.tv_nsec);
            lyr_write_log_record(statep, statep->log_buf, bufsize);    
            #endif  
            data = 1;
        } 
        count++;
    }
	return;
}
    
static int
lyr_strategy(struct buf *bp)
{
	int inst = getminor(bp->b_edev);
	struct lyr_state *statep;
	char *myname = "lyr_strategy";
	statep = (lyr_state_t *)ddi_get_soft_state(lyr_statep, inst);
	if (statep == NULL) {
		cmn_err(CE_WARN, "%s: ddi_get_soft_state failed on "
				"inst %d\n", myname, inst);
		bioerror(bp, ENXIO);
		biodone(bp);
		return (0);
	}
	if (statep->start) {
       if (bp->b_flags & B_WRITE) {
		bp_mapin(bp);
		lyr_process_block_data(bp, statep);
       } else if (bp->b_flags & B_READ) {
        //cmn_err(CE_WARN, "on a read\n");   
        /* This is a block read */
		lyr_process_block_data_onread(bp, statep);
       }
    } 

	return (ldi_strategy(statep->lh, bp));
}

static int
lyr_prop_op(dev_t devt, dev_info_t *dip, ddi_prop_op_t prop_op,
			int flags, char *name, caddr_t valuep, int *lengthp)
{
	int rv, inst = getminor(devt);
	lyr_state_t *statep;
	char *myname = "lyr_prop_op";
	
	statep = (lyr_state_t *)ddi_get_soft_state(lyr_statep, inst);
	if (statep == NULL) {
		cmn_err(CE_WARN, "%s: ddi_get_soft_state failed on inst %d\n",
				myname, inst);
		return (EIO);
	}
	return(ldi_prop_op(statep->lh, prop_op, flags, name, valuep, lengthp)); 
}

void lyr_reset_counters(lyr_state_t *statep) {
    
    statep->dblocks = 0;
    statep->jblocks = 0;
    statep->oblocks = 0;
    statep->ublocks = 0;
    statep->is_start = 1;
    return;
}

void lyr_copy_stats(stats *s, lyr_state_t *statep) {
    
    s->dblocks = statep->dblocks;
    s->jblocks = statep->jblocks;
    s->oblocks = statep->oblocks;
    s->ublocks = statep->ublocks;
    return;
}


static int 
lyr_ioctl(dev_t devt, int cmd, intptr_t arg, int mode,
		  cred_t *credp, int *rvalp)
{
	int rv, inst = getminor(devt);
	lyr_state_t *statep;
	char *myname = "lyr_ioctl";
    stats curr_stats;
	
	statep = (lyr_state_t *)ddi_get_soft_state(lyr_statep, inst);
	if (statep == NULL) {
		cmn_err(CE_WARN, "%s: ddi_get_soft_state failed on inst %d\n",
				myname, inst);
		return (EIO);
	}
    switch (cmd) {

        case LYR_START:
            if (statep->start)
		        cmn_err(CE_WARN, "lyr: already started \n");
            else {    
                size_t bufsize;
                statep->start = 1;

                if (lyr_open_log_file(statep)) 
		            cmn_err(CE_WARN, "XXX: lyr can't open log file \n");
                #ifdef DEBUG_STATS
                if (lyr_open_debug_file(statep)) 
		            cmn_err(CE_WARN, "XXX: lyr can't open debug file \n");
                #endif 
				gethrestime(&statep->curr_time);
                bufsize = snprintf(statep->log_buf, LYR_FS_BLK_SIZE,
                                "%llu\t%llu\t%llu\t%llu\t%lu\t%lu\n", 
                                    statep->oblocks, statep->dblocks, 
                                    statep->jblocks, statep->ublocks/4,
                                    statep->curr_time.tv_sec, 
                                    statep->curr_time.tv_nsec);
                lyr_write_log_record(statep, statep->log_buf, bufsize);
            }
            break;
        case LYR_STOP:
            statep->start = 0;
            lyr_close_log_file(statep);
            #ifdef DEBUG_STATS
                lyr_close_debug_file(statep);
            #endif
            break;
        case LYR_STATS:
            lyr_copy_stats(&curr_stats, statep);
            if (copyout((void *)&curr_stats, (void *)arg, sizeof(stats)))
                return (EFAULT);    
            break;
        case LYR_RESET:
            lyr_reset_counters(statep);
            break;
        default:    
	        return(ldi_ioctl(statep->lh, cmd, arg, mode, credp, rvalp));
	}
	return 0;
}

int lyr_open_log_file(lyr_state_t *statep) {

	int err = vn_open("/tmp/lyr_log", UIO_SYSSPACE, 
                      FWRITE, 0644, &statep->log_vp,
					  0, 0);
	if (err) {
       err = vn_open("/tmp/lyr_log", UIO_SYSSPACE, 
                     FWRITE | FCREAT, 0644, 
                     &statep->log_vp, 0, 0); 
        if (err) {
		    cmn_err(CE_WARN, " *** cant open the log file ***\n");
            goto out;
        } 
   } 

   statep->log_offset = 0;   /* Start from begining of the file */

out:
    return err;
}

int lyr_close_log_file(lyr_state_t *statep) 
{
	int err;
	if (statep->log_vp != NULL) {
		err = fop_close(statep->log_vp, FWRITE, 1, (offset_t)0, 
                        kcred);
		if (err) {
			cmn_err(CE_WARN, "cant close file \n");
		} 
        
		VN_RELE(statep->log_vp);
	}
    return err;
}

int lyr_write_log_record(lyr_state_t *statep, char *buf, int size) {
	int err;
	ssize_t	resid = 0;
    rlim64_t rlimit = RLIM64_INFINITY;
	
    err = vn_rdwr(UIO_WRITE, statep->log_vp, buf, size, statep->log_offset, UIO_SYSSPACE, 
                  FASYNC, rlimit, kcred, &resid);
    
    statep->log_offset += size;
    
    return err;
}

#ifdef DEBUG_STATS

int lyr_open_debug_file(lyr_state_t *statep) {

	int err = vn_open("/tmp/lyr_debug", UIO_SYSSPACE, 
                      FWRITE, 0644, &statep->debug_vp,
					  0, 0);
	if (err) {
       err = vn_open("/tmp/lyr_debug", UIO_SYSSPACE, 
                     FWRITE | FCREAT, 0644, 
                     &statep->debug_vp, 0, 0); 
        if (err) {
		    cmn_err(CE_WARN, " *** cant open the log file ***\n");
            goto out;
        } 
   } 

   statep->debug_offset = 0;   /* Start from begining of the file */

out:
    return err;
}

int lyr_close_debug_file(lyr_state_t *statep) 
{
	int err;
	if (statep->debug_vp != NULL) {
		err = fop_close(statep->debug_vp, FWRITE, 1, (offset_t)0, 
                        kcred);
		if (err) {
			cmn_err(CE_WARN, "cant close file \n");
		} 
        
		VN_RELE(statep->debug_vp);
	}
    return err;
}

int lyr_write_debug_record(lyr_state_t *statep, char *buf, int size) {
	int err;
	ssize_t	resid = 0;
    rlim64_t rlimit = RLIM64_INFINITY;
	
    err = vn_rdwr(UIO_WRITE, statep->debug_vp, buf, size, statep->debug_offset, UIO_SYSSPACE, 
                  FASYNC, rlimit, kcred, &resid);
    
    statep->debug_offset += size;
    
    return err;
}
#endif
