typedef struct dva {
    uint64_t    dva_word[2];
} dva_t;

/*
 * Each block has a 256-bit checksum -- 
   strong enough for cryptographic hashes.
 */
typedef struct zio_cksum {
    uint64_t    zc_word[4];
} zio_cksum_t;

typedef struct blkptr {
    dva_t       blk_dva[3]; /* 128-bit Data Virtual Address */
    uint64_t    blk_prop;   /* size, compression, type, etc */
    uint64_t    blk_pad[3]; /* Extra space for the future   */
    uint64_t    blk_birth;  /* transaction group at birth   */
    uint64_t    blk_fill;   /* fill count           */
    zio_cksum_t blk_cksum;  /* 256-bit checksum     */
} blkptr_t;

#define ZBT_MAGIC   0x210da7ab10c7a11ULL    /* zio data bloc tail */

typedef struct zio_block_tail {
    uint64_t    zbt_magic;  /* for validation, endianness   */
    zio_cksum_t zbt_cksum;  /* 256-bit checksum     */
} zio_block_tail_t;

typedef struct zil_header {
    uint64_t zh_claim_txg;  /* txg in which log blocks were claimed */
    uint64_t zh_replay_seq; /* highest replayed sequence number */
    blkptr_t zh_log;    /* log chain */
    uint64_t zh_claim_seq;  /* highest claimed sequence number */
    uint64_t zh_pad[5];
} zil_header_t;

typedef struct zil_trailer {
    uint64_t zit_pad;
    blkptr_t zit_next_blk;  /* next block in chain */
    uint64_t zit_nused; /* bytes in log block used */
    zio_block_tail_t zit_bt; /* block trailer */
} zil_trailer_t;


typedef enum zil_create {
    Z_FILE,
    Z_DIR,
    Z_XATTRDIR,
} zil_create_t;


/*
 * Intent log transaction types and record structures
 */
#define TX_CREATE       1   /* Create file */
#define TX_MKDIR        2   /* Make directory */
#define TX_MKXATTR      3   /* Make XATTR directory */
#define TX_SYMLINK      4   /* Create symbolic link to a file */
#define TX_REMOVE       5   /* Remove file */
#define TX_RMDIR        6   /* Remove directory */
#define TX_LINK         7   /* Create hard link to a file */
#define TX_RENAME       8   /* Rename a file */
#define TX_WRITE        9   /* File write */
#define TX_TRUNCATE     10  /* Truncate a file */
#define TX_SETATTR      11  /* Set file attributes */
#define TX_ACL_V0       12  /* Set old formatted ACL */
#define TX_ACL          13  /* Set ACL */
#define TX_CREATE_ACL       14  /* create with ACL */
#define TX_CREATE_ATTR      15  /* create + attrs */
#define TX_CREATE_ACL_ATTR  16  /* create with ACL + attrs */
#define TX_MKDIR_ACL        17  /* mkdir with ACL */
#define TX_MKDIR_ATTR       18  /* mkdir with attr */
#define TX_MKDIR_ACL_ATTR   19  /* mkdir with ACL + attrs */
#define TX_MAX_TYPE     20  /* Max transaction type */

typedef struct {            /* common log record header */
    uint64_t    lrc_txtype; /* intent log transaction type */
    uint64_t    lrc_reclen; /* transaction record length */
    uint64_t    lrc_txg;    /* dmu transaction group number */
    uint64_t    lrc_seq;    /* see comment above */
} lr_t;

typedef struct {
    uint32_t    lr_attr_masksize; /* number of elements in array */
    uint32_t    lr_attr_bitmap; /* First entry of array */
    /* remainder of array and any additional fields */
} lr_attr_t;

typedef struct {
    lr_t        lr_common;  /* common portion of log record */
    uint64_t    lr_doid;    /* object id of directory */
    uint64_t    lr_foid;    /* object id of created file object */
    uint64_t    lr_mode;    /* mode of object */
    uint64_t    lr_uid;     /* uid of object */
    uint64_t    lr_gid;     /* gid of object */
    uint64_t    lr_gen;     /* generation (txg of creation) */
    uint64_t    lr_crtime[2];   /* creation time */
    uint64_t    lr_rdev;    /* rdev of object to create */
    /* name of object to create follows this */
    /* for symlinks, link content follows name */
    /* for creates with xvattr data, the name follows the xvattr info */
} lr_create_t;


typedef struct {
    lr_create_t lr_create;  /* common create portion */
    uint64_t    lr_aclcnt;  /* number of ACEs in ACL */
    uint64_t    lr_domcnt;  /* number of unique domains */
    uint64_t    lr_fuidcnt; /* number of real fuids */
    uint64_t    lr_acl_bytes;   /* number of bytes in ACL */
    uint64_t    lr_acl_flags;   /* ACL flags */
    /* lr_acl_bytes number of variable sized ace's follows */
    /* if create is also setting xvattr's, then acl data follows xvattr */
    /* if ACE FUIDs are needed then they will follow the xvattr_t */
    /* Following the FUIDs will be the domain table information. */
    /* The FUIDs for the owner and group will be in the lr_create */
    /* portion of the record. */
    /* name follows ACL data */
} lr_acl_create_t;

typedef struct {
    lr_t        lr_common;  /* common portion of log record */
    uint64_t    lr_doid;    /* obj id of directory */
    /* name of object to remove follows this */
} lr_remove_t;

typedef struct {
    lr_t        lr_common;  /* common portion of log record */
    uint64_t    lr_doid;    /* obj id of directory */
    uint64_t    lr_link_obj;    /* obj id of link */
    /* name of object to link follows this */
} lr_link_t;

typedef struct {
    lr_t        lr_common;  /* common portion of log record */
    uint64_t    lr_sdoid;   /* obj id of source directory */
    uint64_t    lr_tdoid;   /* obj id of target directory */
    /* 2 strings: names of source and destination follow this */
} lr_rename_t;

typedef struct {
    lr_t        lr_common;  /* common portion of log record */
    uint64_t    lr_foid;    /* file object to write */
    uint64_t    lr_offset;  /* offset to write to */
    uint64_t    lr_length;  /* user data length to write */
    uint64_t    lr_blkoff;  /* offset represented by lr_blkptr */
    blkptr_t    lr_blkptr;  /* spa block pointer for replay */
    /* write data will follow for small writes */
} lr_write_t;

typedef struct {
    lr_t        lr_common;  /* common portion of log record */
    uint64_t    lr_foid;    /* object id of file to truncate */
    uint64_t    lr_offset;  /* offset to truncate from */
    uint64_t    lr_length;  /* length to truncate */
} lr_truncate_t;

typedef struct {
    lr_t        lr_common;  /* common portion of log record */
    uint64_t    lr_foid;    /* file object to change attributes */
    uint64_t    lr_mask;    /* mask of attributes to set */
    uint64_t    lr_mode;    /* mode to set */
    uint64_t    lr_uid;     /* uid to set */
    uint64_t    lr_gid;     /* gid to set */
    uint64_t    lr_size;    /* size to set */
    uint64_t    lr_atime[2];    /* access time */
    uint64_t    lr_mtime[2];    /* modification time */
    /* optional attribute lr_attr_t may be here */
} lr_setattr_t;

typedef struct {
    lr_t        lr_common;  /* common portion of log record */
    uint64_t    lr_foid;    /* obj id of file */
    uint64_t    lr_aclcnt;  /* number of acl entries */
    /* lr_aclcnt number of ace_t entries follow this */
} lr_acl_v0_t;

typedef struct {
    lr_t        lr_common;  /* common portion of log record */
    uint64_t    lr_foid;    /* obj id of file */
    uint64_t    lr_aclcnt;  /* number of ACEs in ACL */
    uint64_t    lr_domcnt;  /* number of unique domains */
    uint64_t    lr_fuidcnt; /* number of real fuids */
    uint64_t    lr_acl_bytes;   /* number of bytes in ACL */
    uint64_t    lr_acl_flags;   /* ACL flags */
    /* lr_acl_bytes number of variable sized ace's follows */
} lr_acl_t;

#define SPA_BLKPTRSHIFT 7   /* blkptr_t is 128 bytes    */
#define DNODE_SHIFT     9   /* 512 bytes */
#define DNODE_SIZE  (1 << DNODE_SHIFT)
#define DNODE_CORE_SIZE     64  /* 64 bytes for dnode sans blkptrs */
#define DN_MAX_BONUSLEN (DNODE_SIZE - DNODE_CORE_SIZE - (1 << SPA_BLKPTRSHIFT))

typedef struct dnode_phys {
    uint8_t dn_type;        /* dmu_object_type_t */
    uint8_t dn_indblkshift;     /* ln2(indirect block size) */
    uint8_t dn_nlevels;     /* 1=dn_blkptr->data blocks */
    uint8_t dn_nblkptr;     /* length of dn_blkptr */
    uint8_t dn_bonustype;       /* type of data in bonus buffer */
    uint8_t dn_checksum;        /* ZIO_CHECKSUM type */
    uint8_t dn_compress;        /* ZIO_COMPRESS type */
    uint8_t dn_flags;       /* DNODE_FLAG_* */
    uint16_t dn_datablkszsec;   /* data block size in 512b sectors */
    uint16_t dn_bonuslen;       /* length of dn_bonus */
    uint8_t dn_pad2[4];

    /* accounting is protected by dn_dirty_mtx */
    uint64_t dn_maxblkid;       /* largest allocated block ID */
    uint64_t dn_used;       /* bytes (or sectors) of disk space */

    uint64_t dn_pad3[4];

    blkptr_t dn_blkptr[1];
    uint8_t dn_bonus[DN_MAX_BONUSLEN];
} dnode_phys_t;

typedef struct objset_phys {
    dnode_phys_t os_meta_dnode;
    zil_header_t os_zil_header;
    uint64_t os_type;
    char os_pad[1024 - sizeof (dnode_phys_t) - sizeof (zil_header_t) -
        sizeof (uint64_t)];
} objset_phys_t;


#define UBERBLOCK_MAGIC     0x00bab10c      /* oo-ba-bloc!  */

struct uberblock {
	uint64_t    ub_magic;   /* UBERBLOCK_MAGIC      */
	uint64_t    ub_version; /* SPA_VERSION          */
	uint64_t    ub_txg;     /* txg of last sync     */
	uint64_t    ub_guid_sum;    /* sum of all vdev guids    */
	uint64_t    ub_timestamp;   /* UTC time of last sync    */
	blkptr_t    ub_rootbp;  /* MOS objset_phys_t        */
};

typedef struct uberblock uberblock_t;


#define P2PHASE(x, align)       ((x) & ((align) - 1))
#define BF64_DECODE(x, low, len)    P2PHASE((x) >> (low), 1ULL << (len))
#define BF64_GET(x, low, len)       BF64_DECODE(x, low, len)
#define BF64_GET_SB(x, low, len, shift, bias)   \
    ((BF64_GET(x, low, len) + (bias)) << (shift))


#define SPA_MINBLOCKSHIFT   9

#define DVA_GET_VDEV(dva)   BF64_GET((dva)->dva_word[0], 32, 32)
#define DVA_GET_OFFSET(dva) \
    BF64_GET_SB((dva)->dva_word[1], 0, 63, SPA_MINBLOCKSHIFT, 0)
#define DVA_GET_ASIZE(dva)  \
    BF64_GET_SB((dva)->dva_word[0], 0, 24, SPA_MINBLOCKSHIFT, 0)
#define DVA_GET_GANG(dva)   BF64_GET((dva)->dva_word[1], 63, 1)

